/*
    UHU-menu -- input plugin for freedesktop-style files
    Copyright (C) 2003  Koblinger Egmont <egmont@uhulinux.hu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _GNU_SOURCE
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>

#include "uhumenu.h"

static char *file_content;
static int file_size;

static int path_remove;

static int parse_line (struct item *item, struct section **sp, char *s)
{
	char *lang, *value, *v2, *v3;
	char *end = s + strlen(s);
	while (end > s && (*(end-1) == ' ' || *(end-1) == '\t')) *(--end) = '\0';
	while (*s == ' ' || *s == '\t') s++;
	if (*s == '\0' || *s == '#') return 0;
	if (*s == '[' && *(end-1) == ']') {
		*(--end) = '\0';
		*sp = item_add_section(item, s+1);
		return 0;
	}
	value = strchr(s, '=');
	if (value == NULL) return -1;
	end = value;
	*end = '\0';
	value++;
	while (*value == ' ' || *value == '\t') value++;
	while (end > s && (*(end-1) == ' ' || *(end-1) == '\t')) *(--end) = '\0';
	if (*s == '\0') return -1;
	lang = strchr(s, '[');
	if (lang != NULL && *(end-1) == ']') {
		*lang++ = '\0';
		*(--end) = '\0';
	} else {
		lang = "";
	}
	if (*sp == NULL) {
		message(1, "section == NULL !!!");
		return -1;
	}
	v2 = malloc(strlen(value) + 1);
	v3 = v2;
	while (*value != '\0') {
		if (*value == '\r') {
			value++;
			continue;
		}
		if (*value == '\\') {
			value++;
			switch (*value) {
			  case 's':
				*v3 = ' ';
				break;
			  case 'n':
				*v3 = '\n';
				break;
			  case 't':
				*v3 = '\t';
				break;
			  case 'r':
				*v3 = '\r';
				break;
			  default:
				*v3 = *value;
			}
			if (*v3 != '\0') v3++;
			value++;
		} else {
			*v3++ = *value++;
		}
	}
	*v3 = '\0';
	section_add_value(*sp, s, lang, v2);
	free(v2);
	return 0;
}

static int parse_item (struct item *item)
{
	char *p, *s;
	struct section *section = NULL;

	item->sections = NULL;
	p = file_content;
	while (p < file_content + file_size) {
		s = strchr(p, '\n');
		if (s == NULL) {
			s = p + strlen(p);
		} else {
			*s++ = '\0';
		}
		parse_line(item, &section, p);
		p = s;
	}
	return 0;
}

static int file_to_item (const char *file, struct item *item)
{
	struct stat st;
	int i;
	int fd;

	i = stat(file, &st);
	if (i < 0) {
		message_e(3, "cannot stat %s", file);
		return -1;
	}
	if (S_ISREG(st.st_mode)) {
		item->issubmenu = 0;
	} else {
		message_e(3, "%s is not a regular file", file);
		return -1;
	}

	fd = open(file, O_RDONLY);
	if (fd < 0) {
		message_e(3, "cannot open %s", file);
		return -1;
	}

	if (fstat(fd, &st) < 0) {
		message_e(3, "fstat");
		close(fd);
		return -1;
	}

	item->file_name = strdup(file + path_remove);
	file_size = st.st_size;
	file_content = malloc(file_size + 1);

	if (read(fd, file_content, file_size) != file_size) {
		message(3, "read failed");
		close(fd);
		return -1;
	}
	file_content[file_size] = '\0';

	close(fd);

	return 0;
}

#if 0
static struct menu *build_menu (const char *dir)
{
	DIR *d;
	struct dirent *de;
	char buf[PATH_MAX];
	struct menu *menu;
	int i;

	d = opendir(dir);
	if (d == NULL) {
		message_e(3, "opendir failed");
		return NULL;
	}
	menu = malloc(sizeof(struct menu));
	menu->nr = 0;
	while ((de = readdir_nobackup(d)) != NULL) menu->nr++;
	menu->items = malloc(menu->nr * sizeof(struct item));
	rewinddir(d);
	i = 0;
	while (i < menu->nr && (de = readdir_nobackup(d)) != NULL) {
		sprintf(buf, "%s/%s", dir, de->d_name);
		if (file_to_item(buf, &menu->items[i]) < 0) continue;
		if (parse_item(&menu->items[i]) < 0) continue;
		free(file_content);
		if (menu->items[i].issubmenu) {
			if ((menu->items[i].submenu = build_menu(buf)) == NULL) {
				continue;
			}
		}
		i++;
	}
	menu->nr = i;
	closedir(d);
	if (i == 0) {
		free(menu->items);
		free(menu);
		return NULL;
	}
	return menu;
}
#endif

static struct menu *menu_new (void)
{
	struct menu *menu;
	menu = malloc(sizeof(struct menu));
	menu->nr = 0;
	menu->items = NULL;
	return menu;
}

static int menu_add_item (struct menu *menu, const struct item *item)
{
	if (menu->nr == 0) {
		menu->items = malloc(sizeof(struct item));
	} else {
		menu->items = realloc(menu->items, (menu->nr + 1) * sizeof(struct item));
	}
	if (item != NULL) {
		memcpy(&menu->items[menu->nr], item, sizeof(struct item));
	}
	menu->nr++;
	return 0;
}

static int replace_item (struct item *dst, struct item *src)
{
	src->issubmenu = dst->issubmenu;
	src->submenu = dst->submenu;
	src->file_name = dst->file_name;
	memcpy(dst, src, sizeof(struct item));
	return 0;
}

static int menu_insert_directory_item (struct menu *menu, struct item *item,
  const char *path)
{
	char buf[PATH_MAX];
	int i, j;

	path += strspn(path, "/");

	if (path[0] == '\0') {
		if (menu->nr > 0 && menu->items[0].sections != NULL) {
			message(3, "root menu directory duplicated");
			return -1;
		}
		replace_item(&menu->items[0], item);
		return 0;
	}
	buf[0] = '\0';
	while (1) {
		for (i = 0; i < menu->nr; i++) {
			if (!(strcmp(menu->items[i].file_name, buf))) {
				break;
			}
		}
		if (i == menu->nr) {
			menu_add_item(menu, NULL);
			menu->items[i].issubmenu = 1;
			menu->items[i].submenu = menu_new();
			menu->items[i].file_name = strdup(buf);
			menu->items[i].sections = NULL;
		}
		if (path[0] == '\0') break;
		menu = menu->items[i].submenu;
		j = strcspn(path, "/");
		sprintf(buf + strlen(buf), "/%.*s", j, path);
		path += j;
		path += strspn(path, "/");
	}
	replace_item(&menu->items[i], item);
	return 0;
}

static int menu_insert_file_item (struct menu *menu, struct item *item,
  const char *path)
{
	char buf[PATH_MAX];
	int i, j;

printf("   %s\n", path);
	path += strspn(path, "/");

	buf[0] = '\0';
	while (1) {
		for (i = 0; i < menu->nr; i++) {
printf("%d.. %s\n", i, menu->items[i].file_name);
			if (!(strcmp(menu->items[i].file_name, buf))) {
				break;
			}
		}
printf("asdf\n");
		if (i == menu->nr) {
printf("x\n");
			menu_add_item(menu, NULL);
			menu->items[i].issubmenu = 0;
			menu->items[i].submenu = NULL;
			menu->items[i].file_name = strdup(buf);
			menu->items[i].sections = NULL;
		}
		if (path[0] == '\0') break;
		menu = menu->items[i].submenu;
		j = strcspn(path, "/");
		sprintf(buf + strlen(buf), "/%.*s", j, path);
		path += j;
		path += strspn(path, "/");
	}
	replace_item(&menu->items[i], item);
	return 0;
}

static int read_file (const char *file)
{
	struct item it;

	if (file_to_item(file, &it) < 0) return -1;
	
	return 0;
}

void dump (struct menu *menu, int n)
{
	int i, j;
	if (menu == NULL) return;
	for (i = 0; i < menu->nr; i++) {
		for (j=0; j<2*n; j++) printf(" ");
		printf("\"%s\" %d %p\n", menu->items[i].file_name, menu->items[i].issubmenu, menu->items[i].submenu);
		if (menu->items[i].submenu != NULL) dump(menu->items[i].submenu, n+1);
	}
}

struct menu *read_menu (const char *path)
{
	struct menu *menu;
	struct item it;
	char *fn;
	char fn2[PATH_MAX];
	char buf[PATH_MAX];
	char buf2[PATH_MAX];
	struct dirent *de;
	DIR *d;
	int l;

	menu = menu_new();
	it.issubmenu = 1;
	

	sprintf(buf, "%s/desktop-directories", path);
	d = opendir(buf);
	if (d == NULL) {
		message_e(3, "opendir %s failed", buf);
		return NULL;
	}
	while ((de = readdir(d)) != NULL) {
		l = strlen(de->d_name);
		if (l < 10 || strcmp(de->d_name + l - 10, ".directory")) continue;

		sprintf(buf2, "%s/%s", buf, de->d_name);
printf("%s\n", de->d_name);
		if (file_to_item(buf2, &it) >= 0 && parse_item(&it) >= 0) {
			fn = get_value(&it, NULL, "X-UHU-Menu", "");
			if (fn != NULL) {
				menu_insert_directory_item(menu, &it, fn);
			}
		}
	}
	closedir(d);



	dump(menu, 0);



	sprintf(buf, "%s/applications", path);
	d = opendir(buf);
	if (d == NULL) {
		message_e(3, "opendir %s failed", buf);
		return NULL;
	}
	while ((de = readdir(d)) != NULL) {
		l = strlen(de->d_name);
		if (l < 8 || strcmp(de->d_name + l - 8, ".desktop")) continue;

		sprintf(buf2, "%s/%s", buf, de->d_name);
		if (file_to_item(buf2, &it) >= 0 && parse_item(&it) >= 0) {
			fn = get_value(&it, NULL, "X-UHU-Menu", "");
			if (fn != NULL) {
				sprintf(fn2, "%s/%s", fn, de->d_name);
				menu_insert_file_item(menu, &it, fn2);
			}
		}
	}
	closedir(d);


printf("******************************\n");
	dump(menu, 0);


	return menu;
}
